from tempfile import SpooledTemporaryFile
import requests
import binstar_client
from binstar_client.requests_ext import stream_multipart
from binstar_client.utils import compute_hash, jencode


class ProjectUploader(binstar_client.Binstar):
    def __init__(self, token, **kwargs):
        domain = kwargs.get('domain', 'https://api.anaconda.org')
        verify = kwargs.get('verify', True)
        self.username = kwargs.get('username', None)
        self.project = kwargs.get('project', None)
        super(ProjectUploader, self).__init__(token, domain, verify)

    def exists(self):
        url = "{}/apps/{}/projects/{}".format(
            self.domain, self.username, self.project.name)
        res = self.session.get(url)
        return res.status_code == 200

    def create(self):
        url = "{}/apps/{}/projects".format(self.domain, self.username)
        data, headers = jencode(self.project.to_project_creation())
        res = self.session.post(url, data=data, headers=headers)
        return res

    def stage(self):
        url = "{}/apps/{}/projects/{}/stage".format(
            self.domain, self.username, self.project.name)
        data, headers = jencode(self.project.to_stage())
        res = self.session.post(url, data=data, headers=headers)
        self._check_response(res)
        return res

    def commit(self, revision_id):
        url = "{}/apps/{}/projects/{}/commit/{}".format(
            self.domain, self.username,
            self.project.name, revision_id
        )
        data, headers = jencode({})
        res = self.session.post(url, data=data, headers=headers)
        self._check_response(res, [201])
        return res

    def file_upload(self, url, obj):
        _hexmd5, b64md5, size = compute_hash(
            self.project.tar, size=self.project.size)

        s3data = obj['form_data']
        s3data['Content-Length'] = size
        s3data['Content-MD5'] = b64md5

        data_stream, headers = stream_multipart(
            s3data, files={'file': (self.project.basename, self.project.tar)})

        s3res = requests.post(
            url,
            data=data_stream,
            verify=self.session.verify,
            timeout=10 * 60 * 60,
            headers=headers)

        if s3res.status_code != 201:
            raise binstar_client.errors.BinstarError(
                'Error uploading package', s3res.status_code)
        return s3res

    def projects(self):
        url = "{}/apps/{}/projects".format(self.domain, self.username)
        data, headers = jencode(self.project.to_project_creation())
        return self.session.get(url, data=data, headers=headers)

    def upload(self):
        '''
        * Check if project already exists
            * if it doesn't, then create it
        * stage a new revision
        * upload
        * commit revision
        '''
        if not self.exists():
            self.create()

        data = self.stage().json()
        self.file_upload(data['post_url'], data)
        res = self.commit(data['dist_id'])
        data = res.json()
        return data



'''
Created on Feb 22, 2014

@author: sean
'''
from __future__ import unicode_literals
import requests
from functools import wraps
import json

try:
    unicode
except NameError:
    unicode = str

from collections import namedtuple

rule = namedtuple('rule', ('url', 'path', 'method', 'status', 'content', 'side_effect', 'res', 'headers', 'expected_headers'))

def filter_request(m, prepared_request):
    if m.url and  m.url != prepared_request.url:
        return False

    if m.path and  m.path != prepared_request.path_url:
        return False

    if m.method and m.method != prepared_request.method:
        return False

    return True

class Responses(object):
    def __init__(self):
        self._resps = []

    def append(self, res):
        self._resps.append(res)

    @property
    def called(self):
        return (len(self._resps))

    @property
    def req(self):
        if self._resps:
            return self._resps[0][1]

    def assertCalled(self, url=''):
        assert self.called, "The url %s was not called" % url

    def assertNotCalled(self, url=''):
        assert not self.called, "The url %s was called" % url

class Registry(object):
    def __init__(self):
        self._map = []

    def __enter__(self):
        self.real_send = requests.Session.send
        requests.Session.send = self.mock_send

        return self

    def __exit__(self, *exec_info):
        requests.Session.send = self.real_send
        return

    def mock_send(self, prepared_request, *args, **kwargs):

        rule = next((m for m in self._map[::-1] if filter_request(m, prepared_request)), None)

        if rule is None:
            raise Exception('No matching rule found for url [%s] %s' % (prepared_request.method,
                                                                          prepared_request.url,
                                                                          ))

        if rule.expected_headers:
            for header, value in rule.expected_headers.items():
                if header not in prepared_request.headers:
                    raise Exception("{}: header {} expected in {}".format(prepared_request.url, header, prepared_request.headers))

                if prepared_request.headers[header] != value:
                    raise Exception("{}: header {} has unexpected value {} was expecting {}".format(prepared_request.url, header, prepared_request.headers[header], value))

        content = rule.content
        if isinstance(content, dict):
            content = json.dumps(content)
        if isinstance(content, unicode):
            content = content.encode()

        res = requests.models.Response()
        res.status_code = rule.status
        res._content_consumed = True
        res._content = content
        res.encoding = 'utf-8'
        res.request = prepared_request
        res.headers.update(rule.headers or {})
        rule.res.append((res, prepared_request))

        if rule.side_effect:
            rule.side_effect()

        return res

    def register(self, url=None, path=None, method='GET', status=200, content=b'', side_effect=None, headers=None, expected_headers=None):
        res = Responses()
        self._map.append(rule(url, path, method, status, content, side_effect, res, headers, expected_headers))
        return res

    def unregister(self, res):
        for item in list(self._map):
            if res == item.res:
                self._map.remove(item)
                return

    def assertAllCalled(self):
        for item in self._map:
            res = item.res
            res.assertCalled('[%s] %s%s' % (item.method or 'any', item.url or 'http://<any>', item.path))

def urlpatch(func):
    @wraps(func)
    def inner(self, *args, **kwargs):
        with Registry() as r:
            return func(self, r, *args, **kwargs)
    return inner




'''
Created on May 2, 2014

@author: sean
'''

from binstar_client.utils import jencode
from binstar_client.errors import BinstarError

class ChannelsMixin(object):

    def list_channels(self, owner):
        '''List the channels for owner
        If owner is none, the currently logged in user is used
        '''
        url = '%s/channels/%s' % (self.domain, owner)

        res = self.session.get(url)
        self._check_response(res, [200])
        return res.json()

    def show_channel(self, channel, owner):
        '''List the channels for owner
        If owner is none, the currently logged in user is used
        '''
        url = '%s/channels/%s/%s' % (self.domain, owner, channel)

        res = self.session.get(url)
        self._check_response(res, [200])
        return res.json()

    def add_channel(self, channel, owner, package=None, version=None, filename=None):
        '''
        Add a channel to the specified files
        
        :param channel: channel to add
        :param owner: The user to add the channel to (all files of all packages for this user)
        :param package: The package to add the channel to (all files in this package)
        :param version: The version to add the channel to (all files in this version of the package)
        :param filename: The exact file to add the channel to
        
        '''
        url = '%s/channels/%s/%s' % (self.domain, owner, channel)
        data, headers = jencode(package=package, version=version, basename=filename)

        res = self.session.post(url, data=data, headers=headers)
        self._check_response(res, [201])

    def remove_channel(self, channel, owner, package=None, version=None, filename=None):
        '''
        Remove a channel from the specified files
        
        :param channel: channel to remove
        :param owner: The user to remove the channel from (all files of all packages for this user)
        :param package: The package to remove the channel from (all files in this package)
        :param version: The version to remove the channel to (all files in this version of the package)
        :param filename: The exact file to remove the channel from
        
        '''
        url = '%s/channels/%s/%s' % (self.domain, owner, channel)
        data, headers = jencode(package=package, version=version, basename=filename)

        res = self.session.delete(url, data=data, headers=headers)
        self._check_response(res, [201])

    def copy_channel(self, channel, owner, to_channel):
        '''
        Tag all files in channel <channel> also as channel <to_channel> 
        
        :param channel: channel to copy
        :param owner: Perform this operation on all packages of this user
        :param to_channel: Destination name (may be a channel that already exists)
        
        '''
        url = '%s/channels/%s/%s/copy/%s' % (self.domain, owner, channel, to_channel)
        res = self.session.post(url)
        self._check_response(res, [201])

    def lock_channel(self, channel, owner):
        '''
        Tag all files in channel <channel> also as channel <to_channel> 
        
        :param channel: channel to copy
        :param owner: Perform this operation on all packages of this user
        :param to_channel: Destination name (may be a channel that already exists)
        
        '''
        url = '%s/channels/%s/%s/lock' % (self.domain, owner, channel)
        res = self.session.post(url)
        self._check_response(res, [201])

    def unlock_channel(self, channel, owner):
        '''
        Tag all files in channel <channel> also as channel <to_channel> 
        
        :param channel: channel to copy
        :param owner: Perform this operation on all packages of this user
        :param to_channel: Destination name (may be a channel that already exists)
        
        '''
        url = '%s/channels/%s/%s/lock' % (self.domain, owner, channel)
        res = self.session.delete(url)
        self._check_response(res, [201])




# -*- coding: utf-8 -*-
"""Test anaconda-client configuration set/get."""

# Standard library imports
from os.path import join
import os
import shutil
import tempfile
import unittest

# Third party imports
import mock

# Local imports
from binstar_client.utils import config


class Test(unittest.TestCase):

    def create_config_dirs(self):
        tmpdir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, tmpdir)
        system_dir = join(tmpdir, 'system')
        user_dir = join(tmpdir, 'user')
        os.mkdir(system_dir)
        os.mkdir(user_dir)
        return user_dir, system_dir

    def test_defaults(self):
        user_dir, system_dir = self.create_config_dirs()
        with open(join(user_dir, 'config.yaml'), 'wb') as fd:
            fd.write(b'')

        with mock.patch('binstar_client.utils.config.SEARCH_PATH',
                        [system_dir, user_dir]):
            cfg = config.get_config()
            self.assertEqual(cfg, config.DEFAULT_CONFIG)

    def test_global_url(self):
        """
        Test regression reported on:

        https://github.com/Anaconda-Platform/anaconda-client/issues/464
        """
        user_dir, system_dir = self.create_config_dirs()
        with open(join(user_dir, 'config.yaml'), 'wb') as fd:
            fd.write(b'')

        url_data = {'url': 'https://blob.org'}
        config_data = config.DEFAULT_CONFIG.copy()
        config_data.update(url_data)

        with mock.patch('binstar_client.utils.config.SEARCH_PATH',
                        [system_dir, user_dir]):
            config.save_config(url_data, join(user_dir, 'config.yaml'))
            cfg = config.get_config()
            self.assertEqual(cfg, config_data)

    def test_merge(self):
        user_dir, system_dir = self.create_config_dirs()
        with open(join(system_dir, 'config.yaml'), 'wb') as fd:
            fd.write(b'''
ssl_verify: false
sites:
    develop:
        url: http://develop.anaconda.org
            ''')

        with open(join(user_dir, 'config.yaml'), 'wb') as fd:
            fd.write(b'''
ssl_verify: true
sites:
    develop:
        ssl_verify: false
            ''')

        with mock.patch('binstar_client.utils.config.SEARCH_PATH', [system_dir, user_dir]), \
                mock.patch('binstar_client.utils.config.DEFAULT_CONFIG', {}):
            cfg = config.get_config()

            self.assertEqual(cfg, {
                'ssl_verify': True,
                'sites': {
                    'develop': {
                        'url': 'http://develop.anaconda.org',
                        'ssl_verify': False,
                    }
                }
            })

    def test_support_tags(self):
        user_dir, system_dir = self.create_config_dirs()
        with open(join(user_dir, 'config.yaml'), 'wb') as fd:
            fd.write(b'''
!!python/unicode 'sites':
   !!python/unicode 'alpha': {!!python/unicode 'url': !!python/unicode 'foobar'}
   !!python/unicode 'binstar': {!!python/unicode 'url': !!python/unicode 'barfoo'}
ssl_verify: False
''')

        with mock.patch('binstar_client.utils.config.SEARCH_PATH', [user_dir]), \
                mock.patch('binstar_client.utils.config.DEFAULT_CONFIG', {}):
            cfg = config.get_config()

            self.assertEqual(cfg, {
                'ssl_verify': False,
                'sites': {
                    'alpha': {
                        'url': 'foobar',
                    },
                    'binstar': {
                        'url': 'barfoo',
                    },
                }
            })
